from allencell_ml_segmenter.core.publisher import Publisher
from allencell_ml_segmenter.core.event import Event
from enum import Enum
from typing import Union, Optional, List
from pathlib import Path
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel, ImageType
from qtpy.QtCore import QObject, Signal


class TrainingType(Enum):
    """
    Different cyto-dl experiment types
    """

    SEGMENTATION_PLUGIN = "segmentation_plugin"
    SEGMENTATION = "segmentation"
    GAN = "gan"
    OMNIPOSE = "omnipose"
    SKOOTS = "skoots"


class ModelSize(Enum):
    """
    Model size for training, and their respective filters overrides
    """

    SMALL = [8, 16, 32]
    MEDIUM = [16, 32, 64]
    LARGE = [32, 64, 128]


# TODO: move these signals directly into TrainingModel once we deprecate
# 'Publisher' in a refactor
class TrainingModelSignals(QObject):
    num_channels_set: Signal = Signal()
    images_directory_set: Signal = Signal()


class TrainingModel(Publisher):
    """
    Stores state relevant to training processes.
    """

    def __init__(
        self, main_model: MainModel, experiments_model: ExperimentsModel
    ):
        super().__init__()
        self._main_model = main_model
        self.experiments_model = experiments_model
        self.signals: TrainingModelSignals = TrainingModelSignals()
        self._experiment_type: TrainingType = None
        self._images_directory: Optional[Path] = None
        self._selected_channel: dict[ImageType, Optional[int]] = {
            t: None for t in ImageType
        }
        self._num_channels: dict[ImageType, Optional[int]] = {
            t: None for t in ImageType
        }
        self._model_path: Union[Path, None] = (
            None  # if None, start a new model
        )
        self._patch_size: List[int] = None
        self._spatial_dims: int = None
        self._num_epochs: int = None
        self._max_time: int = None  # in minutes
        self._config_dir: Path = None
        self._max_channel = None
        self._use_max_time: bool = (
            False  # default is false. UI starts with max epoch defined rather than max time.
        )
        self._model_size: Optional[ModelSize] = None
        # the total number of images used for training/test/validation for this model
        self._total_num_images: int = 0

    def get_experiment_type(self) -> Optional[str]:
        """
        Gets experiment type
        """
        if self._experiment_type is None:
            return None
        return self._experiment_type.value

    def set_experiment_type(self, training_type: str) -> None:
        """
        Sets experiment type

        training_type (str): name of cyto-dl experiment to run
        """
        # convert string to enum
        self._experiment_type = TrainingType(training_type)

    def get_spatial_dims(self) -> int:
        """
        Gets image dimensions
        """
        return self._spatial_dims

    def set_spatial_dims(self, spatial_dims: int) -> None:
        """
        Sets image dimensions

        image_dims (int): number of dimensions to train model on. "2" for 2D, "3" for 3D
        """
        if spatial_dims != 2 and spatial_dims != 3:
            raise ValueError("No support for non 2D and 3D images.")
        self._spatial_dims = spatial_dims

    def get_num_epochs(self) -> int:
        """
        Gets max epoch
        """
        return self._num_epochs

    def set_num_epochs(self, num_epochs: int) -> None:
        """
        Sets num epochs

        num_epochs (int): number of additional epochs to train for
        """
        self._num_epochs = num_epochs

    def get_images_directory(self) -> Optional[Path]:
        """
        Gets images directory
        """
        return self._images_directory

    def set_images_directory(self, images_path: Optional[Path]) -> None:
        """
        Sets images directory, and dispatches channel extraction

        images_path (Path): path to images directory
        """
        self._images_directory = images_path
        self.signals.images_directory_set.emit()

    def get_num_channels(self, image_type: ImageType) -> Optional[int]:
        return self._num_channels[image_type]

    def set_all_num_channels(
        self, num_channels: dict[ImageType, Optional[int]]
    ) -> None:
        self._num_channels = num_channels
        self.signals.num_channels_set.emit()

    def get_selected_channel(self, image_type: ImageType) -> Optional[int]:
        return self._selected_channel[image_type]

    def set_selected_channel(
        self, image_type: ImageType, channel: int
    ) -> None:
        self._selected_channel[image_type] = channel

    def get_patch_size(self) -> List[int]:
        """
        Gets patch size
        """
        return self._patch_size

    def set_patch_size(self, patch_size: List[int]) -> None:
        """
        Sets patch size

        patch_size (str): patch size for training
        """
        if len(patch_size) not in [2, 3]:
            raise ValueError(
                "Patch sizes need to be 2 or 3 dimension based on input image."
            )
        self._patch_size = patch_size

    def get_max_time(self) -> int:
        """
        Gets max runtime (in seconds)
        """
        return self._max_time

    def set_max_time(self, max_time: int) -> None:
        """
        Sets max runtime (in seconds)

        max_time (int): maximum runtime for training, in seconds
        """
        self._max_time = max_time

    def get_config_dir(self) -> Path:
        """
        Gets config directory
        """
        return self._config_dir

    def set_config_dir(self, config_dir: Path) -> None:
        """
        Sets config directory

        config_dir (str): path to config directory
        """
        self._config_dir = config_dir

    def dispatch_training(self) -> None:
        """
        Dispatches even to start training
        """
        self.dispatch(Event.PROCESS_TRAINING)

    def use_max_time(self) -> bool:
        """
        Will training run will be based off of max time
        """
        return self._use_max_time

    def set_use_max_time(self, use_max: bool):
        """
        Set if training run will be based off of max time
        """
        self._use_max_time = use_max

    def set_model_size(self, model_size: str) -> None:
        # convert string to enum
        model_size = model_size.upper()
        if model_size not in [x.name for x in ModelSize]:
            raise ValueError(
                "No support for non small, medium, and large patch sizes."
            )
        self._model_size = ModelSize[model_size]

    def get_model_size(self) -> ModelSize:
        return self._model_size

    def set_total_num_images(self, num_images: int) -> None:
        self._total_num_images = num_images

    def get_total_num_images(self) -> int:
        return self._total_num_images

    def get_selected_channels(self) -> dict[ImageType, Optional[int]]:
        return self._main_model.get_selected_channels()



import asyncio

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.aics_widget import AicsWidget
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.sample.sample_state_widget import SampleStateWidget
from allencell_ml_segmenter.sample.sample_results_list_widget import (
    SampleResultsListWidget,
)
from allencell_ml_segmenter.sample.sample_select_files_widget import (
    SampleSelectFilesWidget,
)
from allencell_ml_segmenter.sample.sample_model import SampleModel

from qtpy.QtWidgets import (
    QPushButton,
    QVBoxLayout,
    QSizePolicy,
)

from allencell_ml_segmenter.sample.process.service import SampleProcessService


class SampleView(AicsWidget):
    """
    Sample that orchestrates widgets, managing complex behavior.
    """

    def __init__(self, main_model: MainModel):
        super().__init__()

        # init

        self._main_model = main_model

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.setLayout(QVBoxLayout())

        self._sample_model = SampleModel()
        self._service = SampleProcessService(self._sample_model)

        # children

        self._select_files_widget = SampleSelectFilesWidget(self._sample_model)
        self.layout().addWidget(self._select_files_widget)

        self._btn: QPushButton = QPushButton("Start Training")
        self._btn.clicked.connect(lambda: asyncio.run(self._service.run()))
        self.layout().addWidget(self._btn)

        self.state_widget = SampleStateWidget(self._sample_model)
        layout.addWidget(self.state_widget)

        self._results_list_widget = SampleResultsListWidget(self._sample_model)
        self.layout().addWidget(self._results_list_widget)

        dummy_btn = QPushButton("Dummy Button")
        dummy_btn.clicked.connect(
            lambda: self._main_model.dispatch(Event.PROCESS_TRAINING_COMPLETE)
        )
        self.layout().addWidget(dummy_btn)



from allencell_ml_segmenter.sample.sample_model import SampleModel
from allencell_ml_segmenter.core.aics_widget import AicsWidget
from allencell_ml_segmenter.core.event import Event
from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QLabel,
    QPushButton,
)


class SampleResultsListWidget(AicsWidget):
    """ """

    def __init__(self, model: SampleModel) -> None:
        super().__init__()
        self._model = model

        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._clear_btn = QPushButton("Clear")
        self._clear_btn.clicked.connect(self.clear_training_output_files)

        model.subscribe(
            event=Event.PROCESS_TRAINING_PROGRESS,
            subscriber=self,
            handler=self.update,
        )

    def update(self, event: Event):
        self.clear_training_output_files()
        for file in self._model.get_training_output_files():
            self.layout().addWidget(QLabel(file))
        self.layout().addWidget(QLabel("Training complete!"))
        self.layout().addWidget(self._clear_btn)

    def clear_training_output_files(self):
        for i in reversed(range(self.layout().count())):
            self.layout().itemAt(i).widget().setParent(None)

    def handle_event(self, event: Event) -> None:
        pass



import typing
from qtpy.QtGui import QValidator


# https://doc.qt.io/qtforpython-5/PySide2/QtGui/QValidator.html
class PatchSizeValidator(QValidator):
    # override
    def fixup(self, a0: typing.Optional[str]) -> str:
        """
        This function attempts to change input to be valid according to this validator’s rules.
        It need not result in a valid string: callers of this function must re-test afterwards; the default does nothing
        """
        if a0.isdecimal():
            as_int: int = int(a0)
            # negative and 0 patch sizes not allowed
            if as_int < 4:
                as_int = 4
            # round down to the nearest multiple of 4
            return str(as_int - (as_int % 4))
        else:
            return a0

    # override
    def validate(
        self, a0: typing.Optional[str], a1: int
    ) -> typing.Tuple[QValidator.State, str, int]:
        """
        This virtual function returns Invalid if input is invalid according to this validator’s rules,
        Intermediate if it is likely that a little more editing will make the input acceptable
        (e.g. the user types “4” into a widget which accepts integers between 10 and 99), and Acceptable if the input is valid.

        The function can change both input and pos (the cursor position) if required.
        """
        status: QValidator.State = QValidator.State.Intermediate
        # only if a0 is defined and non-empty can we determine its status
        if a0:
            if a0.isdecimal():
                as_int: int = int(a0)
                if as_int <= 0:
                    status = QValidator.State.Invalid
                elif as_int % 4 == 0:
                    status = QValidator.State.Acceptable
            else:
                status = QValidator.State.Invalid

        return status, a0, a1



import numpy


def set_all_nonzero_values_to(
    image: numpy.ndarray, value: int
) -> numpy.ndarray:
    copy = numpy.copy(image)
    # set all nonzero values in this :image to :value in-place
    copy[copy > 0] = value
    return copy



from abc import ABC, abstractmethod
from typing import List, Optional, Callable
from allencell_ml_segmenter.main.segmenter_layer import (
    ShapesLayer,
    ImageLayer,
    LabelsLayer,
)
import numpy as np
from napari.layers import Layer
from napari.utils.events import Event as NapariEvent


class IViewer(ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def add_image(self, image: np.ndarray, name: str) -> None:
        pass

    @abstractmethod
    def get_image(self, name: str) -> Optional[ImageLayer]:
        pass

    @abstractmethod
    def get_all_images(self) -> List[ImageLayer]:
        pass

    @abstractmethod
    def add_shapes(self, name: str, face_color: str, mode: str) -> None:
        pass

    @abstractmethod
    def get_shapes(self, name: str) -> Optional[ShapesLayer]:
        pass

    @abstractmethod
    def get_all_shapes(self) -> List[ShapesLayer]:
        pass

    @abstractmethod
    def add_labels(self, data: np.ndarray, name: str) -> None:
        pass

    @abstractmethod
    def get_labels(self, name: str) -> Optional[LabelsLayer]:
        pass

    @abstractmethod
    def get_all_labels(self) -> List[LabelsLayer]:
        pass

    @abstractmethod
    def clear_layers(self) -> None:
        pass

    @abstractmethod
    def remove_layer(self, name: str) -> bool:
        pass

    @abstractmethod
    def contains_layer(self, name: str) -> bool:
        pass

    # TODO: refactor prediction/file_input_widget.py to not use this and use get_all_images instead
    @abstractmethod
    def get_layers(self) -> List[Layer]:
        pass

    @abstractmethod
    def subscribe_layers_change_event(
        self, function: Callable[[NapariEvent], None]
    ) -> None:
        pass



from unittest.mock import patch, Mock

import pytest
from qtpy.QtWidgets import QFileDialog
from pytestqt.qtbot import QtBot
from qtpy.QtCore import Qt

from allencell_ml_segmenter.widgets.input_button_widget import InputButton
from allencell_ml_segmenter.prediction.model import PredictionModel


@pytest.fixture
def model_set_file_path_function() -> Mock:
    """
    Fixture that creates a mock function for setting a file path.
    """
    return Mock()


@pytest.fixture
def input_button(
    qtbot: QtBot, model_set_file_path_function: Mock
) -> InputButton:
    """
    Fixture that creates an instance of InputButton for testing. This InputButton is meant to mock selecting files.
    """
    return InputButton(
        Mock(spec=PredictionModel), model_set_file_path_function
    )


def test_set_text(
    qtbot: QtBot, input_button: InputButton, model_set_file_path_function: Mock
) -> None:
    """
    Tests the _update_path_text method of InputButton for file use cases.
    """
    # ARRANGE
    mock_path: str = "/path/to/file"

    with patch.object(
        QFileDialog, "getOpenFileName", return_value=(mock_path, "")
    ):
        # ACT
        qtbot.mouseClick(input_button.button, Qt.LeftButton)

        # ASSERT
        assert input_button._text_display.text() == mock_path
        model_set_file_path_function.assert_called_once()


def test_elongate(input_button: InputButton) -> None:
    """
    Tests the elongate method of InputButton.
    """
    # ARRANGE
    initial_width: int = input_button._text_display.width()
    expected_width: int = initial_width + 100

    # ACT
    input_button.elongate(expected_width)

    # ASSERT
    assert input_button._text_display.width() == expected_width



from pathlib import Path
from typing import List, Optional
from allencell_ml_segmenter.config.i_user_settings import IUserSettings

import copy

from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel


class ExperimentsModel(IExperimentsModel):
    def __init__(self, config: IUserSettings) -> None:
        super().__init__()
        self.user_settings = config

        # options
        self.experiments = []
        self.refresh_experiments()

    def get_checkpoint(self) -> Optional[str]:
        """
        Gets checkpoint
        """
        return self._get_best_ckpt()

    def refresh_experiments(self) -> None:
        # TODO: make a FileUtils method for this?
        self.experiments = []
        for (
            experiment
        ) in self.user_settings.get_user_experiments_path().iterdir():
            if (
                experiment.is_dir()
                and self._is_cyto_dl_experiment(experiment)
                and experiment not in self.experiments
                and not experiment.name.startswith(".")
            ):
                self.experiments.append(experiment.name)
        self.experiments.sort()

    """
    Returns a defensive copy of Experiments list.
    """

    def _is_cyto_dl_experiment(self, experiment: Path) -> bool:
        # Heuristic for checking if dir is a cyto-dl experiment
        csv_path = experiment / "data" / "train.csv"
        checkpoints_path = experiment / "checkpoints"
        return checkpoints_path.exists() or csv_path.exists()

    def get_experiments(self) -> List[str]:
        return copy.deepcopy(self.experiments)

    def get_user_settings(self) -> IUserSettings:
        return self.user_settings

    def get_user_experiments_path(self) -> Path:
        return self.get_user_settings().get_user_experiments_path()

    def get_model_test_images_path(self, experiment_name: str) -> Path:
        return (
            self.get_user_settings().get_user_experiments_path()
            / experiment_name
            / "test_images"
            if experiment_name
            else None
        )

    def get_model_checkpoints_path(
        self, experiment_name: str, checkpoint: str
    ) -> Path:
        """
        Gets checkpoints for model path
        """
        if experiment_name is None:
            raise ValueError(
                "Experiment name cannot be None in order to get model_checkpoint_path"
            )

        if checkpoint is None:
            raise ValueError(
                "Checkpoint cannot be None in order to get model_checkpoint_path"
            )
        return (
            self.get_user_experiments_path()
            / experiment_name
            / "checkpoints"
            / checkpoint
        )

    def get_csv_path(self) -> Optional[Path]:
        if (
            self.get_user_experiments_path() is not None
            and self.get_experiment_name() is not None
        ):
            return (
                self.get_user_experiments_path()
                / self.get_experiment_name()
                / "data"
            )
        return None

    def get_metrics_csv_path(self) -> Path:
        return (
            self.get_user_experiments_path()
            / self.get_experiment_name()
            / "csv"
        )

    def get_cache_dir(self) -> Path:
        return (
            self.get_user_experiments_path()
            / self.get_experiment_name()
            / "cache"
        )

    def get_latest_metrics_csv_version(self) -> int:
        """
        Returns version number of the most recent version directory within
        the cyto-dl CSV folder (self._csv_path) or -1 if no version directories
        exist
        """
        last_version: int = -1
        if self.get_metrics_csv_path().exists():
            for child in self.get_metrics_csv_path().glob("version_*"):
                if child.is_dir():
                    version_str: str = child.name.split("_")[-1]
                    try:
                        last_version = (
                            int(version_str)
                            if int(version_str) > last_version
                            else last_version
                        )
                    except ValueError:
                        continue
        return last_version

    def get_latest_metrics_csv_path(self) -> Optional[Path]:
        version: int = self.get_latest_metrics_csv_version()
        return (
            self.get_metrics_csv_path() / f"version_{version}" / "metrics.csv"
            if version >= 0
            else None
        )

    def get_train_config_path(self, experiment_name: str) -> Path:
        return (
            self.get_user_experiments_path()
            / experiment_name
            / "train_config.yaml"
            if experiment_name
            else None
        )

    def get_current_epoch(self) -> Optional[int]:
        ckpt: Optional[str] = self.get_checkpoint()
        if not ckpt:
            return None
        # assumes checkpoint format: epoch_001.ckpt
        return int(ckpt.split(".")[0].split("_")[-1])

    def _get_best_ckpt(self) -> Optional[str]:
        if not self.get_experiment_name():
            return None

        checkpoints_path = (
            Path(self.user_settings.get_user_experiments_path())
            / self.get_experiment_name()
            / "checkpoints"
        )
        if not checkpoints_path.exists():
            return None

        files: List[Path] = [
            entry
            for entry in checkpoints_path.iterdir()
            if entry.is_file() and not "last" in entry.name.lower()
        ]
        if not files:
            return None

        files.sort(key=lambda file: file.stat().st_mtime)
        return files[-1].name

    def get_channel_selection_path(self) -> Optional[Path]:
        return (
            self.get_csv_path() / "selected_channels.json"
            if self.get_csv_path() is not None
            else None
        )



from abc import abstractmethod
from qtpy.QtCore import QObject, Signal


class ProgressTrackerSignals(QObject):
    progress_changed: Signal = Signal(int)
    label_text_changed: Signal = Signal(str)
    progress_max_changed: Signal = Signal(int)


class ProgressTracker:
    """
    Base class for all ProgressTrackers to inherit from. A ProgressTracker
    maintains an integer measure of progress between progress_minimum and
    progress_maximum. The progress value can be used by PyQt progress bars
    for example. It also maintains a string which can be used as a label for
    the progress.
    """

    def __init__(
        self,
        progress_minimum: int = 0,
        progress_maximum: int = 0,
        label_text: str = "Progress",
    ):
        self._progress_minimum: int = progress_minimum
        self._progress_maximum: int = progress_maximum
        self._progress: int = progress_minimum
        self._label_text: str = label_text
        self.signals: ProgressTrackerSignals = ProgressTrackerSignals()

    def get_progress_minimum(self) -> int:
        return self._progress_minimum

    def get_progress_maximum(self) -> int:
        return self._progress_maximum

    def set_progress_maximum(self, maximum: int) -> None:
        self._progress_maximum = maximum
        self.signals.progress_max_changed.emit(self._progress_maximum)

    def get_progress(self) -> int:
        return self._progress

    def set_progress(self, progress: int) -> None:
        """
        If param progress > progress_maximum, throws ValueError.
        If param progress < progress minimum, throws ValueError.
        Otherwise sets this trackers progress to param progress.
        """
        if progress > self._progress_maximum:
            raise ValueError(
                "cannot set progress to value greater than progress_maximum"
            )
        if progress < self._progress_minimum:
            raise ValueError(
                "cannot set progress to value less than progress_minimum"
            )

        self._progress = progress
        self.signals.progress_changed.emit(self._progress)

    def get_label_text(self) -> str:
        return self._label_text

    def set_label_text(self, label_text: str) -> None:
        self._label_text = label_text
        self.signals.label_text_changed.emit(self._label_text)

    @abstractmethod
    def start_tracker(self) -> None:
        """
        Enables updates to the progress measure from another thread.
        """
        pass

    @abstractmethod
    def stop_tracker(self) -> None:
        """
        Stops any threads that may be active for progress updates.
        Must be called before losing reference to the instance of the ProgressTracker.
        """
        pass



from watchdog.events import FileSystemEvent, FileSystemEventHandler
from pathlib import Path
from csv import DictReader
from typing import Callable, Tuple, Set, Optional


class MetricsCSVEventHandler(FileSystemEventHandler):
    """
    A MetricsCSVEventHandler calls progress_callback upon any changes to
    the provided target_path CSV file, passing the latest epoch in the modified
    CSV to the callback.
    """

    def __init__(
        self,
        target_path: Path,
        progress_callback: Callable,
        label_text_callback: Callable,
    ):
        super().__init__()
        self._target_path: Path = target_path
        self._progress_callback: Callable = progress_callback
        self._label_text_callback: Callable = label_text_callback

    def _get_csv_data(self) -> Tuple[int, Optional[float]]:
        epochs: Set[int] = set()
        latest_loss: Optional[float] = None
        with self._target_path.open("r", newline="") as fr:
            dict_reader: DictReader = DictReader(fr)
            latest_epoch: int = -1
            for row in dict_reader:
                epoch: int = int(row["epoch"])
                epochs.add(epoch)
                if epoch > latest_epoch:
                    try:
                        # this data is not available in all rows for a given epoch
                        latest_loss = float(row["val/loss_epoch"])
                        latest_epoch = epoch
                    except Exception:
                        pass

        return len(epochs), latest_loss

    # override
    def on_any_event(self, event: FileSystemEvent) -> None:
        if self._target_path.exists() and self._target_path.samefile(
            event.src_path
        ):
            epochs, loss = self._get_csv_data()
            self._progress_callback(epochs)
            self._label_text_callback(
                f"Current loss: {'  N/A' if loss is None else '{:.3f}'.format(loss)}"
            )



from .image_data import ImageData
from .i_image_data_extractor import IImageDataExtractor
from .aics_image_data_extractor import AICSImageDataExtractor
from .fake_image_data_extractor import FakeImageDataExtractor



from allencell_ml_segmenter.core.progress_tracker import ProgressTracker
import pytest


def test_set_progress_within_bounds():
    tracker: ProgressTracker = ProgressTracker(
        progress_minimum=0, progress_maximum=10
    )
    tracker.set_progress(2)
    assert tracker.get_progress() == 2
    tracker.set_progress(9)
    assert tracker.get_progress() == 9
    tracker.set_progress(0)
    assert tracker.get_progress() == 0
    tracker.set_progress(10)
    assert tracker.get_progress() == 10


def test_set_progress_greater_than_max():
    tracker: ProgressTracker = ProgressTracker(
        progress_minimum=0, progress_maximum=10
    )
    with pytest.raises(ValueError):
        tracker.set_progress(11)
    with pytest.raises(ValueError):
        tracker.set_progress(10394)


def test_set_progress_less_than_min():
    tracker: ProgressTracker = ProgressTracker(
        progress_minimum=0, progress_maximum=10
    )
    with pytest.raises(ValueError):
        tracker.set_progress(-1)
    with pytest.raises(ValueError):
        tracker.set_progress(-1948)



from qtpy.QtWidgets import QListWidget, QListWidgetItem
from typing import Union, List
from qtpy.QtCore import Qt, Signal


class CheckBoxListWidget(QListWidget):
    """
    A custom QListWidget that allows the user to check/uncheck items in the list.
    Used to display choices for on-screen images in the PredictionView's FileInputWidget.
    """

    checkedSignal: Signal = Signal(int, Qt.CheckState)

    def __init__(self) -> None:
        super().__init__()
        self.setMouseTracking(True)
        self.itemEntered.connect(self._show_tool_tip)
        self.itemChanged.connect(self._send_checked_signal)
        self.setMaximumHeight(100)
        self.setStyleSheet("margin-top: 0px")

    def _send_checked_signal(self, item: QListWidgetItem) -> None:
        """
        Handles the itemChanged signal by emitting a checkedSignal that indicates which item was chosen.
        """
        idx: int = self.row(item)
        self.checkedSignal.emit(idx, item.checkState())

    def add_item(self, item: Union[str, QListWidgetItem]) -> None:
        """
        Adds an item to the list.
        """
        if isinstance(item, str):
            item_add: QListWidgetItem = QListWidgetItem(item)
        elif isinstance(item, QListWidgetItem):
            item_add: QListWidgetItem = item
        else:
            raise TypeError(
                f"Item added to CheckBoxListWidget must be a string or QListWidgetItem, but got {type(item)} instead"
            )

        # set checkable and unchecked by default
        item_add.setFlags(item_add.flags() | Qt.ItemIsUserCheckable)
        item_add.setCheckState(Qt.Unchecked)
        super().addItem(item_add)

    def set_all_state(self, state: Qt.CheckState) -> None:
        for i in range(self.count()):
            item: QListWidgetItem = self.item(i)
            if item.checkState() != state:
                item.setCheckState(state)

    def get_checked_rows(self) -> List[int]:
        return self.__get_flag_rows(Qt.Checked)

    def get_unchecked_rows(self) -> List[int]:
        return self.__get_flag_rows(Qt.Unchecked)

    def __get_flag_rows(self, flag: Qt.CheckState) -> List[int]:
        flag_lst: List[int] = []
        for i in range(self.count()):
            item: QListWidgetItem = self.item(i)
            if item.checkState() == flag:
                flag_lst.append(i)

        return flag_lst

    def remove_checked_rows(self) -> None:
        self.__remove_flag_rows(Qt.Checked)

    def remove_unchecked_rows(self) -> None:
        self.__remove_flag_rows(Qt.Unchecked)

    def __remove_flag_rows(self, flag: Qt.CheckState) -> None:
        flag_lst: List[int] = self.__get_flag_rows(flag)
        for i in reversed(flag_lst):
            self.takeItem(i)

    def _show_tool_tip(self, item: QListWidgetItem) -> None:
        text: str = item.text()
        if self.fontMetrics().boundingRect(text).width() > self.width():
            self.setToolTip(text)
        else:
            self.setToolTip("")



__version__ = "0.1.16rc2"

from allencell_ml_segmenter.napari.napari_reader import napari_get_reader
from allencell_ml_segmenter.napari.sample_data import make_sample_data
from allencell_ml_segmenter.napari.napari_writer import (
    write_multiple,
    write_single_image,
)

from allencell_ml_segmenter.main.main_widget import MainWidget


__all__ = (
    "napari_get_reader",
    "write_single_image",
    "write_multiple",
    "make_sample_data",
    "MainWidget",
)



from dataclasses import dataclass
import numpy as np
from pathlib import Path
from typing import Optional


@dataclass
class SegmenterLayer:
    name: str


@dataclass
class ShapesLayer(SegmenterLayer):
    data: np.ndarray


@dataclass
class ImageLayer(SegmenterLayer):
    path: Optional[Path]


@dataclass
class LabelsLayer(SegmenterLayer):
    pass



from pathlib import Path
from typing import Dict

from allencell_ml_segmenter.core.directories import Directories


class Style:
    """
    Helper class that enables the retrieval of custom Qt stylesheets. Implementation adapted from Segmenter Classic.
    """

    cache: Dict[str, str] = dict()

    @classmethod
    def get_stylesheet(cls, name: str) -> str:
        """
        Retrieve a stylesheet from the style directory. Stylesheets are cached in memory for efficiency.
        """
        if name is None:
            raise ValueError("Stylesheet name can't be None")
        if not name.endswith(".qss"):
            raise ValueError("Stylesheet must be a qss file (.qss)")

        if name not in cls.cache:
            cls.cache[name] = cls._load_from_file(name)

        return cls.cache[name]

    @classmethod
    def _load_from_file(cls, name: str) -> str:
        """
        Helper method that loads a stylesheet that has not yet been cached.
        """
        path: Path = Directories.get_style_dir() / name
        with open(path, "r") as handle:
            return handle.read()



from pathlib import Path
from typing import Optional

import numpy as np
from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    ImageData,
)


class FakeImageDataExtractor(IImageDataExtractor):
    """
    Returns dummy image data for testing.
    """

    _instance = None

    def extract_image_data(
        self,
        img_path: Path,
        channel: int = 0,
        dims: bool = True,
        np_data: bool = True,
        seg: Optional[int] = None,
    ) -> ImageData:
        return ImageData(
            1 if dims else None,
            2 if dims else None,
            3 if dims else None,
            4 if dims else None,
            np.zeros((5, 5)) if np_data else None,
            img_path,
        )

    @classmethod
    def global_instance(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from pathlib import Path
from typing import Set

import pytest
from pytestqt.qtbot import QtBot
from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.config.i_user_settings import IUserSettings

from allencell_ml_segmenter.core.aics_widget import AicsWidget
from allencell_ml_segmenter.main.main_widget import MainWidget
from unittest.mock import Mock
import napari

# IMPORTANT NOTE: MainWidget is different from the other widgets since we do not directly
# instantiate it in our code. So, it will always receive a napari.Viewer object in
# production. We cannot initialize with our FakeViewer because our 'Viewer' is created during
# initialization of MainWidget. We could supply a "viewer factory" to the MainWidget,
# but for now I'm just mocking it here.


@pytest.fixture
def main_widget(qtbot: QtBot) -> MainWidget:
    """
    Returns a MainWidget instance for testing.
    """
    settings: IUserSettings = FakeUserSettings()
    settings.set_cyto_dl_home_path(Path())
    settings.set_user_experiments_path(Path())
    return MainWidget(viewer=Mock(), settings=settings)


def test_tabs_react_to_new_model_event(
    main_widget: MainWidget,
) -> None:
    """
    Tests that the main widget handles the action new model event correctly.
    """

    # ACT: have the model dispatch the action new model event
    main_widget._model.set_new_model(True)

    # ASSERT: check that the main widget's current view (after setting) is curation
    assert (
        main_widget._window_container.currentIndex()
        == main_widget._window_to_index[main_widget._curation_view]
    )
    # ASSERT: check that the correct tabs are enabled
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._prediction_view]
        )
        == True
    )
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._curation_view]
        )
        == True
    )
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._training_view]
        )
        == True
    )


def test_tabs_react_to_existing_model_event(
    main_widget: MainWidget,
) -> None:
    """
    Tests that the main widget handles the action new model event correctly.
    """

    # ACT: have the model dispatch the action new model event
    main_widget._model.set_new_model(False)

    # ASSERT: check that the main widget's current view (after setting) is prediction
    assert (
        main_widget._window_container.currentIndex()
        == main_widget._window_to_index[main_widget._prediction_view]
    )
    # ASSERT: check that the correct tabs are enabled
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._prediction_view]
        )
        == True
    )
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._curation_view]
        )
        == False
    )
    assert (
        main_widget._window_container.isTabEnabled(
            main_widget._window_to_index[main_widget._training_view]
        )
        == False
    )


def test_handle_action_change_view_event(
    main_widget: MainWidget,
) -> None:
    """
    Tests that the main widget handles the action change view event correctly.
    """
    # ARRANGE
    views: Set[AicsWidget] = main_widget._window_to_index.keys()

    for view in views:
        # ACT: have the model dispatch the action change view event
        main_widget._model.set_current_view(view)

        # ASSERT: check that the main widget's current view (after setting) is same as the model's current view
        assert (
            main_widget._window_container.currentIndex()
            == main_widget._window_to_index[view]
        )


def test_experiments_home_initialized(qtbot: QtBot) -> None:
    """
    Tests that the main widget promtps the user to choose a. 'experiments home' dir if one is not found in user settings.
    """
    # ARRANGE
    EXPECTED_EXPERIMENTS_HOME = Path(
        __file__
    ).parent  # simulates (in the fake settings) the location chosed by user
    settings = FakeUserSettings(
        init_prompt_response=EXPECTED_EXPERIMENTS_HOME,
        cyto_dl_home_path=Path("foo/cyto/path"),
    )
    settings.set_user_experiments_path(
        None
    )  # Simulates state where users has not yet chosen an experiments home.

    # ACT
    MainWidget(
        Mock(spec=napari.Viewer), settings
    )  # If the users settings does not find an experiments home path, it will prompt the user for one and persist it.

    # ASSERT
    assert (
        settings.get_user_experiments_path() == EXPECTED_EXPERIMENTS_HOME
    )  # The path chosen by the user should have been persisted in settings.


def test_tab_enabled(main_widget) -> None:
    """
    Tests that the main widget enables the correct tabs when the experiment is applied.
    """

    # ARRANGE
    main_widget._experiments_model.apply_experiment_name("foo")

    # Sanity check
    assert main_widget._window_container.isEnabled() == True

    # ACT
    main_widget._experiments_model.apply_experiment_name(None)

    # ASSERT
    assert main_widget._window_container.isEnabled() == False

    # ACT
    main_widget._experiments_model.apply_experiment_name("foo")

    # Sanity check
    assert main_widget._window_container.isEnabled() == True



from pathlib import Path
import allencell_ml_segmenter
from allencell_ml_segmenter.training.metrics_csv_event_handler import (
    MetricsCSVEventHandler,
)
from unittest.mock import Mock


def test_csv_3_epochs():
    callback_mock: Mock = Mock()
    label_mock: Mock = Mock()
    test_csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
        / "1_exp"
        / "csv"
        / "version_1"
        / "test_metrics_csv_3_epochs.csv"
    )
    handler: MetricsCSVEventHandler = MetricsCSVEventHandler(
        test_csv_path, callback_mock, label_mock
    )
    fs_event_mock: Mock = Mock(src_path=test_csv_path)
    handler.on_any_event(fs_event_mock)
    callback_mock.assert_called_with(3)


def test_empty_csv():
    callback_mock: Mock = Mock()
    label_mock: Mock = Mock()
    test_csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
        / "1_exp"
        / "csv"
        / "version_0"
        / "test_metrics_csv_empty.csv"
    )
    handler: MetricsCSVEventHandler = MetricsCSVEventHandler(
        test_csv_path, callback_mock, label_mock
    )
    fs_event_mock: Mock = Mock(src_path=test_csv_path)
    handler.on_any_event(fs_event_mock)
    # since test_metrics_csv_empty.csv is empty, we expect 0
    callback_mock.assert_called_with(0)


def test_nonexistent_csv():
    callback_mock: Mock = Mock()
    label_mock: Mock = Mock()
    test_csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
        / "0_exp"
        / "csv"
        / "version_0"
        / "test_metrics_does_not_exist.csv"
    )
    handler: MetricsCSVEventHandler = MetricsCSVEventHandler(
        test_csv_path, callback_mock, label_mock
    )
    fs_event_mock: Mock = Mock(src_path=test_csv_path)
    handler.on_any_event(fs_event_mock)
    callback_mock.assert_not_called()



# This file is intended to be called by a github workflow
import subprocess
import toml


def main():
    version: str = toml.load("pyproject.toml")["project"]["version"]
    tag_output: subprocess.CompletedProcess = subprocess.run(
        ["git", "tag", f"v{version}"]
    )
    if tag_output.returncode != 0:
        raise RuntimeError("failed to tag")


if __name__ == "__main__":
    main()



from typing import Optional

from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QComboBox,
    QVBoxLayout,
    QPushButton,
)

from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.main.experiments_model import IExperimentsModel
from allencell_ml_segmenter.utils.s3.s3_model_bucket import (
    S3ModelBucket,
)
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import PROD_BUCKET


class ModelDownloadDialog(QDialog):
    def __init__(
        self, parent: Optional[QWidget], experiments_model: IExperimentsModel
    ):
        super().__init__(parent)
        self._experiments_model = experiments_model
        self._available_models = S3ModelBucket(
            PROD_BUCKET, self._experiments_model.get_user_experiments_path()
        ).get_available_models()
        self.setLayout(QVBoxLayout())

        self._model_select_dropdown: QComboBox = QComboBox()
        self._model_select_dropdown.setCurrentIndex(-1)
        self._model_select_dropdown.addItems(self._available_models.keys())
        self.layout().addWidget(self._model_select_dropdown)

        self._download_button: QPushButton = QPushButton("Download")
        self._download_button.clicked.connect(self._download_button_handler)
        self.layout().addWidget(self._download_button)

    def _download_button_handler(self) -> None:
        selected_model_name: str = str(
            self._model_select_dropdown.currentText()
        )

        self._available_models[selected_model_name].download_model_and_unzip()
        download_complete_message = InfoDialogBox(
            f"Downloaded {selected_model_name} to {self._experiments_model.get_user_experiments_path() / selected_model_name}"
        )
        download_complete_message.exec()
        self._experiments_model.refresh_experiments()  # prevents repeat downloads of model before exiting download dialog



from .i_file_writer import IFileWriter
import numpy as np
from pathlib import Path
import csv
import json
from io import TextIOBase
from typing import List, Dict, Tuple, Union


class FileWriter(IFileWriter):
    _instance = None

    # maps path to an open csv file and the corresponding csv writer
    _open_files: Dict[Path, Tuple[TextIOBase, csv.writer]] = {}

    def np_save(self, path: Path, arr: np.ndarray) -> None:
        """
        Saves :param arr: to :param path:. Creates directories as necessary.
        """
        path.parent.mkdir(parents=True, exist_ok=True)
        np.save(path, arr)

    def csv_open_write_mode(self, path: Path) -> None:
        """
        Opens a CSV at :param path: in write mode. Can only call csv_write_row
        on the path after running this method. Must call csv_close with the same
        path at some point after calling this method. Creates directories as necessary.
        """
        path = path.resolve()
        if path in self._open_files:
            raise RuntimeError(f"{path} already open")
        path.parent.mkdir(parents=True, exist_ok=True)
        open_file: TextIOBase = open(path, "w", newline="")
        writer: csv.writer = csv.writer(open_file, delimiter=",")
        self._open_files[path] = (open_file, writer)

    def csv_write_row(self, path: Path, row: List[str]) -> None:
        """
        Appends :param row: to the open CSV at :param path:
        """
        path = path.resolve()
        if path not in self._open_files:
            raise RuntimeError(
                f"{path} must be opened with csv_open_write_mode before writing"
            )
        self._open_files[path][1].writerow(row)

    def csv_close(self, path: Path) -> None:
        """
        Closes the open CSV at :param path:
        """
        path = path.resolve()
        if path in self._open_files:
            self._open_files[path][0].close()
            del self._open_files[path]

    def write_json(self, json_like_obj: Union[list, dict], path: Path) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as fw:
            json.dump(json_like_obj, fw)

    @classmethod
    def global_instance(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from unittest.mock import Mock
from dataclasses import dataclass
from pathlib import Path

import pytest

from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationView,
    ImageType,
)
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.curation.curation_widget import CurationWidget
import allencell_ml_segmenter
from allencell_ml_segmenter.main.main_model import MainModel


IMG_DIR_PATH = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "img_folder"
)

IMG_DIR_FILES = [path for path in IMG_DIR_PATH.iterdir()]


@dataclass
class TestEnvironment:
    model: CurationModel
    widget: CurationWidget


@pytest.fixture
def test_env() -> TestEnvironment:
    model: CurationModel = CurationModel(FakeExperimentsModel(), MainModel())
    return TestEnvironment(model, CurationWidget(FakeViewer(), model))


def test_view_change(qtbot: QtBot, test_env: TestEnvironment) -> None:
    # Arrange
    test_env.model.set_image_directory_paths(ImageType.RAW, IMG_DIR_FILES)
    test_env.model.set_image_directory_paths(ImageType.SEG1, IMG_DIR_FILES)
    test_env.model.set_image_directory_paths(ImageType.SEG2, IMG_DIR_FILES)

    # Act / Assert
    assert test_env.widget.get_view() == CurationView.INPUT_VIEW
    test_env.model.set_current_view(CurationView.MAIN_VIEW)
    assert test_env.widget.get_view() == CurationView.MAIN_VIEW



from pathlib import Path

from napari.layers import Layer, Shapes, Image, Labels
from napari.layers.shapes.shapes import Mode
from napari.utils.events import Event as NapariEvent

from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.main.segmenter_layer import (
    ShapesLayer,
    ImageLayer,
    LabelsLayer,
)
import napari
from typing import List, Callable, Optional
import numpy as np


class Viewer(IViewer):
    def __init__(
        self,
        viewer: napari.Viewer,
    ):
        super().__init__()
        self.viewer: napari.Viewer = viewer

    def add_image(self, image: np.ndarray, name: str) -> None:
        self.viewer.add_image(image, name=name)

    def get_image(self, name: str) -> Optional[ImageLayer]:
        for img in self.get_all_images():
            if img.name == name:
                return img
        return None

    def get_all_images(self) -> List[ImageLayer]:
        imgs: List[ImageLayer] = []
        for l in self.viewer.layers:
            if isinstance(l, Image) and l.source.path:
                imgs.append(ImageLayer(l.name, Path(l.source.path)))
            elif isinstance(l, Image):
                imgs.append(ImageLayer(l.name, None))
        return imgs

    def add_shapes(self, name: str, face_color: str, mode: Mode) -> None:
        shapes: Shapes = self.viewer.add_shapes(
            None, name=name, face_color=face_color
        )
        shapes.mode = mode

    def get_shapes(self, name: str) -> Optional[ShapesLayer]:
        for shapes in self.get_all_shapes():
            if shapes.name == name:
                return shapes
        return None

    def get_all_shapes(self) -> List[ShapesLayer]:
        return [
            ShapesLayer(l.name, np.asarray(l.data, dtype=object))
            for l in self.viewer.layers
            if isinstance(l, Shapes)
        ]

    def add_labels(self, data: np.ndarray, name: str) -> None:
        self.viewer.add_labels(data, name=name)

    def get_labels(self, name: str) -> Optional[LabelsLayer]:
        for labels in self.get_all_labels():
            if labels.name == name:
                return labels
        return None

    def get_all_labels(self) -> List[LabelsLayer]:
        # all items in self.viewer.layers inherit from napari.layers.Layer
        # possible types outlined in https://napari.org/stable/api/napari.layers.html
        return [
            LabelsLayer(l.name)
            for l in self.viewer.layers
            if isinstance(l, Labels)
        ]

    def clear_layers(self) -> None:
        self.viewer.layers.clear()

    def remove_layer(self, name: str) -> bool:
        layer: Optional[Layer] = self._get_layer_by_name(name)
        if layer is not None:
            self.viewer.layers.remove(layer)
            return True
        return False

    def contains_layer(self, name: str) -> bool:
        return self._get_layer_by_name(name) is not None

    def get_layers(self) -> List[Layer]:
        return [l for l in self.viewer.layers]

    def subscribe_layers_change_event(
        self, function: Callable[[NapariEvent], None]
    ) -> None:
        self.viewer.events.layers_change.connect(function)

    def _get_layer_by_name(self, name: str) -> Optional[Layer]:
        layers: List[Layer] = self.get_layers()
        for l in layers:
            if l.name == name:
                return l
        return None



from allencell_ml_segmenter.prediction.prediction_folder_event_handler import (
    PredictionFolderEventHandler,
)
from unittest.mock import Mock


def test_file_creation_omesegpredtif():
    # ARRANGE
    fake_file_path: str = "/path/to/some.ome_seg_pred.tif"
    progress_callback_mock: Mock = Mock()
    fs_file_created_event_mock: Mock = Mock(src_path=fake_file_path)
    handler: PredictionFolderEventHandler = PredictionFolderEventHandler(
        progress_callback_mock
    )

    # ACT/ASSERT
    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(1)

    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(2)

    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(3)


def test_file_creation_omesegpredtiff():
    # ARRANGE
    fake_file_path: str = "/path/to/some.ome_seg_pred.tiff"
    progress_callback_mock: Mock = Mock()
    fs_file_created_event_mock: Mock = Mock(src_path=fake_file_path)
    handler: PredictionFolderEventHandler = PredictionFolderEventHandler(
        progress_callback_mock
    )

    # ACT/ASSERT
    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(1)

    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(2)

    handler.on_created(fs_file_created_event_mock)
    progress_callback_mock.assert_called_with(3)


def test_file_creation_bad_ext():
    # ARRANGE
    progress_callback_mock: Mock = Mock()
    handler: PredictionFolderEventHandler = PredictionFolderEventHandler(
        progress_callback_mock
    )

    # ACT/ASSERT
    handler.on_created(Mock(src_path="/bad/file/path.png"))
    progress_callback_mock.assert_not_called()

    handler.on_created(Mock(src_path="/bad/file/.DS_Store"))
    progress_callback_mock.assert_not_called()

    handler.on_created(Mock(src_path="/bad/file/path.zip"))
    progress_callback_mock.assert_not_called()



import pytest
from unittest.mock import MagicMock
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QWidget,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QPushButton,
)
from allencell_ml_segmenter.core.dialog_box import DialogBox
from pytestqt.qtbot import QtBot
from qtpy.QtCore import Qt


@pytest.fixture
def dialog(qtbot: QtBot):
    dialog = DialogBox("Test Message")
    yield dialog


def test_init(dialog):
    assert dialog.getMessage() == "Test Message"


def test_set_message(dialog):
    dialog.setMessage("New Message")
    assert dialog.getMessage() == "New Message"


def test_yes_selected(dialog):
    dialog.yes_selected()
    assert dialog.selection is True


def test_no_selected(dialog):
    dialog.no_selected()
    assert dialog.selection is False


def test_button_connections(dialog, qtbot: QtBot):
    # Manually triggering the click event for testing
    qtbot.mouseClick(
        dialog.yes_btn,
        Qt.LeftButton,
    )
    assert dialog.selection is True

    qtbot.mouseClick(
        dialog.no_btn,
        Qt.LeftButton,
    )
    assert dialog.selection is False



import numpy as np

from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.main.segmenter_layer import (
    ShapesLayer,
    ImageLayer,
    LabelsLayer,
)


from napari.utils.events import Event as NapariEvent
import napari
from typing import List, Dict, Callable, Optional
from napari.layers import Layer


class FakeNapariEvent:
    pass


class FakeViewer(IViewer):
    def __init__(self, viewer: Optional[napari.Viewer] = None):
        self._image_layers: Dict[str, ImageLayer] = {}
        self._shapes_layers: Dict[str, ShapesLayer] = {}
        self._labels_layers: Dict[str, LabelsLayer] = {}
        self._on_layers_change_fns: List[Callable] = []

    def add_image(self, image: np.ndarray, name: str):
        self._image_layers[name] = ImageLayer(name, None)
        self._on_layers_change()

    def get_image(self, name: str) -> Optional[ImageLayer]:
        if name in self._image_layers:
            return self._image_layers[name]
        return None

    def get_all_images(self) -> List[ImageLayer]:
        return [v for k, v in self._image_layers.items()]

    def add_shapes(self, name: str, face_color: str, mode: str) -> None:
        self._shapes_layers[name] = ShapesLayer(
            name, np.asarray([[1, 2], [3, 4]])
        )
        self._on_layers_change()

    def get_shapes(self, name: str) -> Optional[ShapesLayer]:
        if name in self._shapes_layers:
            return self._shapes_layers[name]
        return None

    def get_all_shapes(self) -> List[ShapesLayer]:
        return [v for k, v in self._shapes_layers.items()]

    # NOTE: this method should only exist in fake viewer to simulate drawing shapes
    def modify_shapes(self, name: str, new_shapes: np.ndarray) -> None:
        if name in self._shapes_layers:
            self._shapes_layers[name] = ShapesLayer(name, new_shapes)

    def add_labels(self, data: np.ndarray, name: str) -> None:
        self._labels_layers[name] = LabelsLayer(name)
        self._on_layers_change()

    def get_labels(self, name: str) -> Optional[LabelsLayer]:
        if name in self._labels_layers:
            return self._labels_layers[name]
        return None

    def get_all_labels(self) -> List[LabelsLayer]:
        return [v for k, v in self._labels_layers.items()]

    def clear_layers(self) -> None:
        self._image_layers = {}
        self._shapes_layers = {}

    def remove_layer(self, name: str) -> bool:
        removed: bool = False
        if name in self._image_layers:
            del self._image_layers[name]
            removed = True
        if name in self._shapes_layers:
            del self._shapes_layers[name]
            removed = True
        if removed:
            self._on_layers_change()
        return removed

    def contains_layer(self, name: str) -> bool:
        return (
            name in self._image_layers
            or name in self._shapes_layers
            or name in self._labels_layers
        )

    # not supporting in the fake because we will move away from this fn in the near future
    def get_layers(self) -> List[Layer]:
        return []

    def subscribe_layers_change_event(
        self, function: Callable[[NapariEvent], None]
    ) -> None:
        self._on_layers_change_fns.append(function)

    def _on_layers_change(self):
        for fn in self._on_layers_change_fns:
            fn(FakeNapariEvent())



from allencell_ml_segmenter.core.q_runnable_manager import IQRunnableManager
from qtpy.QtCore import QRunnable, QThreadPool


class GlobalQRunnableManager(IQRunnableManager):
    """
    Executes QRunnables with the global instance of QThreadPool
    """

    _instance = None

    def run(self, runnable: QRunnable) -> None:
        QThreadPool.globalInstance().start(runnable)

    @classmethod
    def global_instance(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from dataclasses import dataclass
from pathlib import Path
import numpy as np


@dataclass
class ImageData:
    dim_x: int
    dim_y: int
    dim_z: int
    channels: int
    np_data: np.ndarray
    path: Path



from allencell_ml_segmenter.main.main_service import MainService
from allencell_ml_segmenter.main.main_model import MainModel, ImageType
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.utils.file_writer import FakeFileWriter
from allencell_ml_segmenter.core.task_executor import SynchroTaskExecutor
from typing import Optional
from pathlib import Path
import allencell_ml_segmenter


# raw: 5, seg1: 2, seg2: 1
mixed_channel_path: Path = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "channel_selection_json"
    / "valid_mixed.json"
)

nonexistent_channel_path: Path = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "channel_selection_json"
    / "DNE.json"
)


def test_service_sets_channels_on_exp_applied_mixed():
    # Arrange
    main_model: MainModel = MainModel()
    exp_model: FakeExperimentsModel = FakeExperimentsModel(
        channel_selection_path=mixed_channel_path
    )
    service: MainService = MainService(
        main_model,
        exp_model,
        task_executor=SynchroTaskExecutor.global_instance(),
        file_writer=FakeFileWriter(),
    )

    # Assert (sanity check)
    channels: dict[ImageType, Optional[int]] = (
        main_model.get_selected_channels()
    )
    assert channels[ImageType.RAW] is None
    assert channels[ImageType.SEG1] is None
    assert channels[ImageType.SEG2] is None

    # Act
    exp_model.apply_experiment_name("test")

    # Assert
    channels = main_model.get_selected_channels()
    assert channels[ImageType.RAW] == 5
    assert channels[ImageType.SEG1] == 2
    assert channels[ImageType.SEG2] == 1


def test_service_does_not_set_channels_on_exp_applied_nonexistent():
    # Arrange
    main_model: MainModel = MainModel()
    exp_model: FakeExperimentsModel = FakeExperimentsModel(
        channel_selection_path=nonexistent_channel_path
    )
    service: MainService = MainService(
        main_model,
        exp_model,
        task_executor=SynchroTaskExecutor.global_instance(),
        file_writer=FakeFileWriter(),
    )

    # Assert (sanity check)
    channels: dict[ImageType, Optional[int]] = (
        main_model.get_selected_channels()
    )
    assert channels[ImageType.RAW] is None
    assert channels[ImageType.SEG1] is None
    assert channels[ImageType.SEG2] is None

    # Act
    exp_model.apply_experiment_name("test")

    # Assert
    channels = main_model.get_selected_channels()
    # since the JSON is not available, channels should all be None still
    assert channels[ImageType.RAW] is None
    assert channels[ImageType.SEG1] is None
    assert channels[ImageType.SEG2] is None


def test_service_writes_to_disk_when_channels_change():
    # Arrange
    main_model: MainModel = MainModel()
    exp_model: FakeExperimentsModel = FakeExperimentsModel(
        channel_selection_path=nonexistent_channel_path
    )
    file_writer: FakeFileWriter = FakeFileWriter()
    service: MainService = MainService(
        main_model,
        exp_model,
        task_executor=SynchroTaskExecutor.global_instance(),
        file_writer=file_writer,
    )

    # Assert (sanity check)
    channels: dict[ImageType, Optional[int]] = (
        main_model.get_selected_channels()
    )
    assert channels[ImageType.RAW] is None
    assert channels[ImageType.SEG1] is None
    assert channels[ImageType.SEG2] is None
    assert len(file_writer.json_state) == 0

    # Act
    main_model.set_selected_channels(
        {
            ImageType.RAW: 8,
            ImageType.SEG1: 2,
            ImageType.SEG2: 7,
        }
    )

    # Assert
    assert len(file_writer.json_state) == 1
    assert file_writer.json_state[nonexistent_channel_path] == {
        "raw": 8,
        "seg1": 2,
        "seg2": 7,
    }



from abc import ABC, abstractmethod
from pathlib import Path
from qtpy.QtWidgets import QWidget


class IUserSettings(ABC):
    def __init__(self):
        super.__init__(self)

    @abstractmethod
    def get_cyto_dl_home_path(self) -> Path:
        pass

    @abstractmethod
    def get_user_experiments_path(self) -> Path:
        pass

    @abstractmethod
    def set_user_experiments_path(self, path: Path):
        pass

    @abstractmethod
    def prompt_for_user_experiments_home(self, parent: QWidget):
        pass

    @abstractmethod
    def display_change_user_experiments_home(self, parent: QWidget):
        pass



from allencell_ml_segmenter.core.publisher import Publisher
from allencell_ml_segmenter.core.event import Event


class SampleModel(Publisher):
    """
    Sample model for this application
    """

    def __init__(self):
        super().__init__()
        self._error_message: str = ""
        self._process_running: bool = False
        self._training_input_files: list = []
        self._training_output_files: list = []

    def get_process_running(self) -> bool:
        """
        getter/property for process running
        """
        return self._process_running

    def get_training_input_files(self) -> list:
        """
        getter/property for training input files
        """
        return self._training_input_files

    def get_training_output_files(self) -> list:
        """
        getter/property for training output files
        """
        return self._training_output_files

    def get_error_message(self) -> str:
        """
        getter/property for error message
        """
        return self._error_message

    # Methods for setting state and dispatching events#
    ###################################################

    def set_error_message(self, error_message: str):
        """
        Set the error message in the UI and dispatch a MainEvent
        """
        self._error_message = error_message
        self.dispatch(Event.PROCESS_TRAINING_SHOW_ERROR)

    def set_process_running(self, process_running: bool):
        """
        Set the process running in the UI and dispatch a MainEvent
        """
        self._process_running = process_running
        self.dispatch(Event.PROCESS_TRAINING_CLEAR_ERROR)
        self.dispatch(Event.PROCESS_TRAINING)

    def append_training_output_files(self, training_output_files: list):
        """
        Append to the training output files in the UI and dispatch a MainEvent
        """
        self._training_output_files.extend(training_output_files)
        self.dispatch(Event.PROCESS_TRAINING_CLEAR_ERROR)
        self.dispatch(Event.PROCESS_TRAINING_PROGRESS)

    def append_training_input_files(self, training_input_files: list):
        """
        Append to the training input files in the UI and dispatch a MainEvent
        """
        self._training_input_files.extend(training_input_files)
        self.dispatch(Event.PROCESS_TRAINING_CLEAR_ERROR)

    def set_training_input_files(self, training_input_files: list):
        """
        Set the training input files in the UI and dispatch a MainEvent
        """
        self._training_input_files = training_input_files
        self.dispatch(Event.PROCESS_TRAINING_CLEAR_ERROR)
